import dataclasses
import hashlib


def size_prefixed(_bytes: bytes) -> bytes:
    return len(_bytes).to_bytes(4, 'big') + _bytes


@dataclasses.dataclass(frozen=True)
class Address:
    _bytes: bytes

    def __post_init__(self):
        assert len(self._bytes) == 21, len(self._bytes)

    def rpc_serialize(self) -> bytes:
        return self._bytes

    def __str__(self) -> str:
        return self._bytes.hex()

    @staticmethod
    def from_string(s: 'Address | str') -> 'Address':
        if isinstance(s, Address):
            return s
        return Address(bytes.fromhex(s))


@dataclasses.dataclass(frozen=True)
class Signature:
    _bytes: bytes

    def __post_init__(self):
        assert len(self._bytes) == 65, len(self._bytes)
        assert self._bytes[0] < 4, 'First byte is the recovery id. Must be less than 4'

    def rpc_serialize(self) -> bytes:
        return self._bytes

    def __str__(self) -> str:
        return self._bytes.hex()


@dataclasses.dataclass(frozen=True)
class HashSha256:
    _bytes: bytes

    def __post_init__(self):
        assert len(self._bytes) == 32, len(self._bytes)

    @staticmethod
    def of_bytes(b: bytes) -> 'HashSha256':
        return HashSha256(hashlib.sha256(b).digest())

    def rpc_serialize(self) -> bytes:
        return self._bytes

    def __str__(self) -> str:
        return self._bytes.hex()


@dataclasses.dataclass(frozen=True)
class Transaction:
    contract_address: Address
    transaction_rpc: bytes

    def __post_init__(self):
        assert isinstance(self.contract_address, Address), self.contract_address

    def rpc_serialize(self) -> bytes:
        return self.contract_address.rpc_serialize() + size_prefixed(
            self.transaction_rpc,
        )


@dataclasses.dataclass(frozen=True)
class SignedTransactionInnerPart:
    nonce: int
    valid_to_time: int
    gas_cost: int
    transaction: Transaction

    def rpc_serialize(self) -> bytes:
        return (
            self.nonce.to_bytes(8, 'big')
            + self.valid_to_time.to_bytes(8, 'big')
            + self.gas_cost.to_bytes(8, 'big')
            + self.transaction.rpc_serialize()
        )

    def hash(self, chain_id: str) -> HashSha256:
        return HashSha256.of_bytes(
            self.rpc_serialize() + size_prefixed(chain_id.encode('utf8')),
        )


@dataclasses.dataclass(frozen=True)
class SignedTransaction:
    inner: SignedTransactionInnerPart
    signature: Signature

    def hash(self, chain_id: str) -> HashSha256:
        return self.inner.hash(chain_id)

    def rpc_serialize(self) -> bytes:
        return self.signature.rpc_serialize() + self.inner.rpc_serialize()
